from __future__ import absolute_import, division, print_function, unicode_literals
from future import standard_library
standard_library.install_aliases()
from builtins import input
from builtins import object
import sys

def default_upload_progress_callback(encoder):
  pass

def progressbar_callback(monitor):
  total_bytes = monitor.encoder.len
  pct = float(monitor.bytes_read) / total_bytes

  l_fill = "{:<27}".format("=" * min(int((min(0.5, pct) / 0.5) / 0.5 * 30), 27))
  p_fill = "{:^4}".format(str(int(pct * 100)) + "%")
  r_fill = "{:<27}".format("=" * min(int(min(pct - 0.5, 0.5) / 0.5 * 30), 27))
  ratio = "{!s}/{!s}".format(monitor.bytes_read, total_bytes)
  line = "[{} {} {}] {}".format(l_fill, p_fill, r_fill, ratio)
  sys.stdout.write("\r{}".format(line))
  sys.stdout.flush()


from __future__ import absolute_import, division, print_function, unicode_literals
from future import standard_library
standard_library.install_aliases()
from builtins import input
from builtins import object
import sys
import os
import time
import requests
import json
import shutil
import copy
import zipfile
from itertools import cycle
from requests_toolbelt import MultipartEncoder, MultipartEncoderMonitor
from .uploadcallbacks import progressbar_callback
from .sessionbuilder import SessionBuilder, ProjectAssistantAuthenticationError


def default_app_data_dir():
  APPNAME = "udacity-pa"
  if sys.platform == 'win32':
    return os.path.join(os.environ['APPDATA'], APPNAME)
  else:
    return os.path.expanduser(os.path.join("~", "." + APPNAME))

class ProjectAssistantSubmissionError(Exception):
  def __init__(self, value):
    self.value = value
  def __str__(self):
    return self.value

def result_filename(project_name, submission_id):
  return "%s-result-%s.json" % (project_name, submission_id)

def submission_filename(project_name, submission_id):
  return "%s-%s.zip" % (project_name, submission_id)

PRESUBMIT_FILENAME = 'student.zip'

def root_url(environment):
  url = {'development': 'http://local-dev.udacity.com:3000',
         'staging': 'https://project-assistant-staging.udacity.com',
         'production': 'https://project-assistant.udacity.com'}

  return url[environment]

def build_session(environment = 'production', jwt_path = None):
  return SessionBuilder(root_url(environment), jwt_path).new()

#Zipfile helper function
def mkzip(root_path, zipfilename, filenames, max_zip_size):
  abs_root_path = os.path.abspath(root_path)
  abspaths = [os.path.abspath(x) for x in filenames]

  if os.path.commonprefix([abs_root_path] + abspaths) != abs_root_path:
    raise ProjectAssistantSubmissionError("Submitted files must in subdirectories of %s." % (root_path or "./"))

  with zipfile.ZipFile(zipfilename,'w') as z:
    for f in filenames:
      zpath = os.path.relpath(f, root_path)
      z.write(f, zpath)

  if os.stat(zipfilename).st_size > max_zip_size:
    raise ProjectAssistantSubmissionError("Your zipfile exceeded the limit of %d bytes" % max_zip_size)

def submit(nanodegree, 
           project, 
           filenames,
           environment = 'production', 
           max_zip_size = 8 << 20,
           jwt_path = None,
           refresh_time = 3):

    try:
      session = build_session(environment, jwt_path)
    except ProjectAssistantAuthenticationError as e:
      print(e)
      return
    
    submission = Submission(nanodegree,
                            project,
                            session,
                            filenames,
                            max_zip_size = max_zip_size,
                            upload_progress_callback = progressbar_callback,
                            environment = environment)

    print("Submission includes the following files:")
    print('\n'.join(['    ' + f for f in submission.filenames]))
    print("")

    print("Uploading submission...")
    try:
      submission.submit()
    except ProjectAssistantSubmissionError as e:
      print('\n' + str(e))
      return

    print('\n')
    wheel = cycle(['|', '/', '-', '\\'])
    spin_freq = 8.
    while not submission.poll():
      for _ in range(int(refresh_time * spin_freq)):
        sys.stdout.write("\rWaiting for results... {}".format(next(wheel)))
        sys.stdout.flush()
        time.sleep(1. / spin_freq)
    sys.stdout.write("\rWaiting for results...Done!\n\n")

    print("Results:\n--------")
    if submission.feedback():
      if submission.console():
        print(submission.console())

      rfilename = result_filename(submission.project_name(), submission.id())
      sfilename = submission_filename(submission.project_name(), submission.id())

      with open(rfilename, "w") as fd:
        json.dump(submission.feedback(), fd, indent=4, separators=(',', ': '))

      print("\nDetails are available in %s.\n" % rfilename)
      print("If you would like this version of the project to be reviewed,\n" \
            "submit %s to the reviews website.\n" %  (sfilename))

    elif submission.error_report():
        print(json.dumps(submission.error_report(), indent=4))
        print("For help troubleshooting, please see the FAQ:\n https://project-assistant.udacity.com/faq")
    else:
        print("Unknown error.")


#Submissions for Nanodegrees
class Submission(object):
  def __init__(self, 
               nanodegree, 
               project,
               session,
               filenames,
               max_zip_size = 8 << 20,
               upload_progress_callback = None,
               environment = 'production'):

    self.nanodegree = nanodegree
    self.project = project
    self.environment = environment

    self.s = session
    self.filenames = copy.deepcopy(filenames)
    self.max_zip_size = max_zip_size
    self.upload_progress_callback = upload_progress_callback or default_upload_progress_callback

  def project_name(self):
    return self.project

  def _get_submit_url(self):
    return root_url(self.environment) + "/student/nanodegree/%s/project/%s/submission" % (self.nanodegree, self.project)   

  def _get_poll_url(self):
    return root_url(self.environment) + "/student/submissions/%s" % (self.id())

  def submit(self):

    self.submit_url = self._get_submit_url()

    mkzip(os.getcwd(), PRESUBMIT_FILENAME, self.filenames, self.max_zip_size)

    fd = open(PRESUBMIT_FILENAME, "rb")

    m = MultipartEncoder(fields={'zipfile': ('student.zip', fd, 'application/zip')})
    monitor = MultipartEncoderMonitor(m, self.upload_progress_callback)

    try:
      r = self.s.post(self.submit_url, 
                      data=monitor,
                      headers={'Content-Type': monitor.content_type})
      r.raise_for_status()
    except requests.exceptions.HTTPError as e:
      if r.status_code == 403:
        raise ProjectAssistantSubmissionError("You don't have access to this project.")
      elif r.status_code in [404,429,500]:
        try:
          response_json = r.json()
          message = response_json.get("message") or "An internal server error occurred."
        except:
          message = "An unknown error occurred"
        raise ProjectAssistantSubmissionError(message)
      else:
        raise

    fd.close()

    self.submission = r.json()

    shutil.move(PRESUBMIT_FILENAME,
                submission_filename(self.project_name(), self.id()))

  def poll(self):
    r = self.s.get(self._get_poll_url())
    r.raise_for_status()

    self.submission = r.json()

    return self.submission['feedback'] is not None or self.submission['error_report'] is not None

  def result(self):
    return self.feedback()

  def feedback(self):
    return self.submission['feedback']

  def console(self):
    return self.submission['console']

  def error_report(self):
    return self.submission['error_report']

  def id(self):
    return self.submission['id']



from __future__ import absolute_import, division, print_function, unicode_literals
from future import standard_library
from future.utils import bytes_to_native_str
standard_library.install_aliases()
from builtins import input
from builtins import object

import os
import sys
import json
import re
import getpass
import errno
import copy
import requests
from urllib.parse import urlsplit

HOTH_URL = "https://hoth.udacity.com"

class ProjectAssistantAuthenticationError(Exception):
  def __init__(self, value):
    self.value = value
  def __str__(self):
    return self.value

def default_app_data_dir():
  APPNAME = "udacity-pa"
  if sys.platform == 'win32':
    return os.path.join(os.environ['APPDATA'], APPNAME)
  else:
    return os.path.expanduser(os.path.join("~", "." + APPNAME))

def default_jwt_paths():
  return [os.path.join(os.getcwd(), 'jwt'), os.path.join(default_app_data_dir(), 'jwt')]

class SessionBuilder():
  def __init__(self, root_url, jwt_path):
    self.root_url = root_url
    self.jwt_path = jwt_path

    if self.jwt_path is None:
      for p in default_jwt_paths():
        if os.path.isfile(p):
          self.jwt_path = p
          break

  def new(self):
    session = requests.Session()
    session.headers.update({'content-type':'application/json', 'accept': 'application/json'})

    jwt = self.load_jwt_from_file()

    if jwt is None or not self.jwt_works(jwt):
      jwt = self.login_for_jwt()

      if jwt is None or not self.jwt_works(jwt):
        raise ProjectAssistantAuthenticationError("Authentication Failed.")

      save = input("Save login token to %s?[y,N]" % self.jwt_path)
      if save.lower() == 'y':
        self.save_the_jwt(jwt)

    self.set_auth_headers(session, jwt)

    return session

  def set_auth_headers(self, session, jwt):
    session.headers.update({'authorization': 'Bearer ' + jwt})

  def jwt_works(self, jwt):
    session = requests.Session()
    session.headers.update({'content-type':'application/json', 'accept': 'application/json'})

    self.set_auth_headers(session, jwt)

    r = session.get(url = self.root_url + '/users/me')

    return r.status_code == 200

  def save_the_jwt(self, jwt):
    try:
      os.makedirs(os.path.dirname(self.jwt_path))
    except OSError as exception:
      if exception.errno != errno.EEXIST:
        raise

    with open(self.jwt_path, "w") as fd:
      fd.write(jwt)


  def load_jwt_from_file(self):
    try:
      with open(self.jwt_path, "r") as fd:
        jwt = fd.read().rstrip()

      if not jwt or not self.jwt_works(jwt):
        jwt = None
    except (requests.exceptions.HTTPError, IOError, ValueError, KeyError) as e:
      jwt = None

    return jwt

  def login_for_jwt(self):
    try:
      session = requests.Session()
      session.headers.update({'content-type':'application/json', 'accept': 'application/json'})

      password_prompt = bytes_to_native_str(b"Password :")

      print("Udacity Login required.")
      email = input('Email :')
      password = getpass.getpass(password_prompt)
      udacity_login(session, self.root_url, email, password)
    except requests.exceptions.HTTPError as e:
      if e.response.status_code == 403:
        raise ProjectAssistantAuthenticationError("Authentication failed")
      else:
        raise e

    try:
      r = session.post(self.root_url + '/auth_tokens')
      r.raise_for_status()
    except:
      raise ProjectAssistantAuthenticationError("Authentication failed")

    jwt = r.json()['auth_token']

    return jwt

#Helper functions for logins
def udacity_login(http, root_url, email, password):
  data = {'email' : email, 'password' : password, "next": root_url + "/auth/udacity/callback"}

  #Logging into udacity
  r = http.post(HOTH_URL + '/v2/authenticate', 
                data=data,
                headers={"content-type": "application/x-www-form-urlencoded", "accept": "*/*"})
  r.raise_for_status()



from __future__ import absolute_import, division, print_function, unicode_literals
from future import standard_library
standard_library.install_aliases()
from builtins import input
from builtins import object

import os
import sys
import argparse
import errno
import requests
import json
from urllib.request import getproxies
from io import StringIO

from pkg_resources import Requirement, resource_filename

from .udacity import build_session, root_url, result_filename, submission_filename
from .sessionbuilder import default_jwt_paths

class NotProjectAssistantError(Exception):
  """Exception raised when not an Udacity Project Assistant directory."""

  def __init__(self):
    super(NotProjectAssistantError, self).__init__("Not an Udacity Project Assistant directory. No .udacity-pa folder found.")

pa_dir = os.path.join(os.path.realpath(os.getcwd()),'.udacity-pa')
if not os.path.isfile(os.path.join(pa_dir,'projects.py')):
  raise  NotProjectAssistantError()

try:
  sys.path.append(pa_dir)
  import projects
  projects.nanodegree
  projects.projects
  projects.submit
except:
  raise ValueError(".udacity-pa/projects.py file is invalid.")

class ActionHelper(object):
  def __init__(self, args):
    self.environment = args.environment
    self.jwt_path = args.jwt_path
    self.root_url = root_url(self.environment)

  def build_session(self):
    return build_session(self.environment, self.jwt_path)

class GetHelper(ActionHelper):
  def __init__(self, args):
    super(GetHelper, self).__init__(args)

    self.submission_id = args.submission_id

  def get_url(self):
    return self.root_url + "/student/submissions/%s" % self.submission_id

  def act(self):
    http = self.build_session()

    url = self.get_url()

    r = http.get(url)
    r.raise_for_status()

    submission = r.json()

    zipfile_url = submission['zipfile']['url']
    project_name = submission['project_name']

    #Downloading the zipfile
    local_zip_filename = submission_filename(project_name, submission['id'])
    r = requests.get(zipfile_url, stream=True)
    with open(local_zip_filename, 'wb') as f:
        for chunk in r.iter_content(chunk_size=1024): 
            if chunk:
                f.write(chunk)
    
    print(submission['console'])

    local_result_filename = result_filename(project_name, submission['id'])

    with open(local_result_filename, "w") as fd:
      try:
        json.dump(submission['feedback'], fd, indent=4, separators=(',', ': '))
      except:
        fd.write(submission['feedback'])

    print("\nDetails available in %s.\n" % local_result_filename)
    print("If you would like this version of the project to be reviewed,\n" \
          "submit %s to the reviews website.\n" %  local_zip_filename)


class LSHelper(ActionHelper):
  def __init__(self, args, ndkey):
    super(LSHelper, self).__init__(args)

    self.ndkey = ndkey
    self.project_name = args.project_name

  def ls_url(self):
    return self.root_url + "/student/nanodegree/%s/project/%s" % (self.ndkey, self.project_name)

  def act(self):
    http = self.build_session()

    url = self.ls_url()

    r = http.get(url)
    r.raise_for_status()

    output = StringIO()
    date_header = '{:40s}'.format('Submission Time')
    id_header = 'ID'.rjust(9)
    output.write('%s %s' % (date_header, id_header))
    output.write('\n' + '-' * 50)
    for s in r.json():
      date = '{:40s}'.format(s['created_at'][:39])
      s_id = ("%d" % s['id']).rjust(9)
      output.write('\n%s %s' % (date, s_id))

    print(output.getvalue())

class SubmitHelper(ActionHelper):
  def __init__(self, args):
    super(SubmitHelper, self).__init__(args)

    self.args = args

  def act(self):
    projects.submit(self.args)

class DiagnoseHelper(object):
  def __init__(self, args):
    self.args = args

  def act(self):
    data = {
      'proxies' : getproxies(),
      'default_jwt_paths': default_jwt_paths(),
      'jwt_present': any(os.path.isfile(jwt_path) for jwt_path in default_jwt_paths()),
      'projects_file_present': os.path.isfile(os.path.join(os.path.realpath(os.getcwd()),
                                                    '.udacity-pa',
                                                    'projects.py')),
      'terminal': os.environ.get('TERM'),
      'platform': sys.platform
    }
    json.dump(data, sys.stdout, indent=4)

def main(args, nanodegree):
  if args.action == 'ls':
    return LSHelper(args, nanodegree).act()
  elif args.action == 'get':
    return GetHelper(args).act()
  elif args.action == 'submit':
    return SubmitHelper(args).act()
  elif args.action == 'diagnose':
    return DiagnoseHelper(args).act()

def main_func():

  parser = argparse.ArgumentParser(description='CLI for submitting projects to the udacity project assistant')

  parser.add_argument('--environment', default='production', help="webserver environment")
  parser.add_argument('--jwt_path', default=None, help="path to file containing auth information")

  subparsers = parser.add_subparsers(dest="action", help="Action")

  ls_parser = subparsers.add_parser("ls")
  ls_parser.add_argument('project_name', 
                         choices = projects.projects, 
                         default = projects.projects[0],
                         nargs = '?',
                         help="project name")

  get_parser = subparsers.add_parser("get")
  get_parser.add_argument('submission_id', help="submission_id")

  submit_parser = subparsers.add_parser("submit")
  submit_parser.add_argument('args', nargs=argparse.REMAINDER)

  diagnose_parser = subparsers.add_parser("diagnose")
  diagnose_parser.add_argument('args', nargs=argparse.REMAINDER)

  args = parser.parse_args()

  main(args, projects.nanodegree)

  return 0



from setuptools import setup

setup(
    name='udacity-pa',
    version='0.2.7',
    author='S. Charles Brubaker',
    author_email='cb@udacity.com',
    packages=['udacity_pa'],
    entry_points={
        'console_scripts': [
            'udacity = udacity_pa.projectassistant:main_func',
            'udacity_pa = udacity_pa.projectassistant:main_func'
        ],
    },
    include_package_data=True,
    url='http://github.com/udacity/udacity-pa',
    license='MIT',
    description="CLI tool for Udacity's project assistant.",
    keywords = 'Udacity',
    long_description=open('README.rst').read(),
    install_requires=[
        "requests >= 2.2.1",
        "requests-toolbelt >= 0.7.0",
        "future"
    ],
)
